/* ====================================================================
 *  Copyright 2003-2004 Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 * ====================================================================
 */

/**
 * @author Paul Querna <chip force-elite.com>
 *
 *  Originaly Based on mod_mysql_pool by Nick Kew:
 *    http://apache.webthing.com/
 *
 * $Id: mod_dbi_pool.c 209 2004-06-19 22:35:50Z chip $
 */

#include "mod_dbi_pool_private.h"
#include "mod_dbi_pool.h"
typedef struct dbi_config dbi_config;
struct dbi_config
{
    const char name;
    dbi_config_rec rec;
#ifdef APR_HAS_THREADS
    apr_reslist_t *pool;
#else
    dbi_conn *conn;
#endif
};


module AP_MODULE_DECLARE_DATA dbi_pool_module;

static apr_hash_t *dbi_pool_config;

dbi_config *create_new_conf(conn_id conn_id, apr_pool_t * p)
{
    dbi_config *conf;

    conf = (dbi_config *) apr_pcalloc(p, sizeof(dbi_config));
    if (conf == NULL) {
        return NULL;
    }
    conf->rec.dbi_dbname = DFLT_DBI_DBNAME;
    conf->rec.dbi_driver = DFLT_DBI_DRIVER;
    conf->rec.dbi_host = DFLT_DBI_HOST;
    conf->rec.dbi_user = DFLT_DBI_USER;
    conf->rec.dbi_pass = DFLT_DBI_PASS;
    conf->rec.conn_min = DFLT_CONN_MIN;
    conf->rec.conn_soft = DFLT_CONN_SOFT;
    conf->rec.conn_max = DFLT_CONN_MAX;
    conf->rec.conn_ttl = DFLT_CONN_TTL;
    apr_hash_set(dbi_pool_config, conn_id, APR_HASH_KEY_STRING, conf);
    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p,
                  "[mod_dbi_pool.c] Creating Config for %s", conn_id);
    return conf;
}

static apr_status_t
get_or_create_dbi_conf(const char *conn_id,
                       apr_pool_t * p, dbi_config ** confname)
{
    dbi_config *temp;
    unsigned int c;

    /*
     * some sanity checks on conn_id: limits are liberal and are more or less random 
     */
    if (strlen(conn_id) > 255) {
        return APR_EGENERAL;
    }
    for (c = 0; c < strlen(conn_id); c++) {
        if (conn_id[c] < ' ') {
            return APR_EGENERAL;
        }
    }
    temp = apr_hash_get(dbi_pool_config, conn_id, APR_HASH_KEY_STRING);
    if (temp == NULL) {
        /*
         * no such server yet... 
         */
        temp = create_new_conf(conn_id, p);
    }
    *confname = temp;
    return APR_SUCCESS;
}

static const char *set_dbi_switch_conf(cmd_parms * cmd, void *config,
                                       const char *conn_id, const char *value)
{
    apr_ssize_t pos = (apr_ssize_t) cmd->info;

    dbi_config *temp;

    if ((get_or_create_dbi_conf(conn_id, cmd->pool, &temp)) == APR_SUCCESS) {

        /*
         * Overwriting an existing value technically is a memory leak, since the 
         * pconf pool is only destroyed at the termination of the whole apache 
         * process. Otoh, when processing htaccess, we get handed the 
         * request-pool instead which is freed afterwards, so we should be fine. 
         */
        /* XXXX: Since mod_pool_dbi will be around alot longer, this needs to be fixed! */

        switch (pos) {
        case CONF_DBI_DRIVER:
            temp->rec.dbi_driver = value;
            break;
        case CONF_DBI_HOST:
            temp->rec.dbi_host = value;
            break;
        case CONF_DBI_USERNAME:
            temp->rec.dbi_user = value;
            break;
        case CONF_DBI_PASSWORD:
            temp->rec.dbi_pass = value;
            break;
        case CONF_DBI_DBNAME:
            temp->rec.dbi_dbname = value;
            break;
        case CONF_DBI_CONN_MIN:
            temp->rec.conn_min = atoi(value);
            break;
        case CONF_DBI_CONN_SOFTMAX:
            temp->rec.conn_soft = atoi(value);
            break;
        case CONF_DBI_CONN_HARDMAX:
            temp->rec.conn_max = atoi(value);
            break;
        case CONF_DBI_CONN_TTL:
            temp->rec.conn_ttl = atoi(value);
            break;
        default:
            // unknown config directive?
            break;
        }
    }
    return NULL;
}

static const command_rec dbi_pool_cmds[] = {
    /*
     * global config items 
     */
    AP_INIT_TAKE2("PoolDbiDriver", set_dbi_switch_conf,
                  (void *) CONF_DBI_DRIVER, RSRC_CONF,
                  "The DBI Driver"),
    AP_INIT_TAKE2("PoolDbiHost", set_dbi_switch_conf,
                  (void *) CONF_DBI_HOST,
                  RSRC_CONF,
                  "The host for the database connection"),
    AP_INIT_TAKE2("PoolDbiUsername", set_dbi_switch_conf,
                  (void *) CONF_DBI_USERNAME, RSRC_CONF,
                  "The username for the database connection"),
    AP_INIT_TAKE2("PoolDbiPassword", set_dbi_switch_conf,
                  (void *) CONF_DBI_PASSWORD, RSRC_CONF,
                  "The password for the database connection"),
    AP_INIT_TAKE2("PoolDbiDBName", set_dbi_switch_conf,
                  (void *) CONF_DBI_DBNAME,
                  RSRC_CONF,
                  "The name of the database containing the tables"),
    AP_INIT_TAKE2("PoolDbiConnMin", set_dbi_switch_conf,
                  (void *) CONF_DBI_CONN_MIN, RSRC_CONF,
                  "The Minimum Number of Database Connections"),
    AP_INIT_TAKE2("PoolDbiConnSoftMax", set_dbi_switch_conf,
                  (void *) CONF_DBI_CONN_SOFTMAX, RSRC_CONF,
                  "The Soft Maximum Number of Database Connections"),
    AP_INIT_TAKE2("PoolDbiConnHardMax", set_dbi_switch_conf,
                  (void *) CONF_DBI_CONN_HARDMAX, RSRC_CONF,
                  "The Hard Maximum Number of Database Connections"),
    AP_INIT_TAKE2("PoolDbiConnTTL", set_dbi_switch_conf,
                  (void *) CONF_DBI_CONN_TTL, RSRC_CONF,
                  "The Database Pool Time To Live for Each Connection."),
    {NULL}
};




/* an apr_reslist_constructor for DBI connections */
static apr_status_t
dbipool_construct(void **db, void *params, apr_pool_t * pool)
{
    dbi_config_rec *conf = params;
    int err_num = 0;
    const char *err_str;

    *db = dbi_conn_new(conf->dbi_driver);

    // Test for a Fatal Error on Creating the Driver
    if (*db == NULL) {
        ap_log_perror(APLOG_MARK, APLOG_EMERG, 0, pool,
                      "[mod_dbi_pool.c] DBI Connection Failed. dbi_conn_new returned NULL. Insufficient memory or invalid DBD?");
            /**
         * modules/ssl/ssl_engine_log.c:103
         *  said this was okay. so i do it too!
         */
        exit(1);
    }

    dbi_conn_set_option(*db, "host", conf->dbi_host);
    dbi_conn_set_option(*db, "username", conf->dbi_user);
    dbi_conn_set_option(*db, "password", conf->dbi_pass);
    dbi_conn_set_option(*db, "dbname", conf->dbi_dbname);

    if (dbi_conn_connect(*db) != 0) {
        err_num = dbi_conn_error(*db, &err_str);
        /*
         * Connetion Failed 
         */
        ap_log_perror(APLOG_MARK, APLOG_CRIT, 0, pool,
                      "[mod_dbi_pool.c] DBI Connection to %s://%s@%s/%s Failed. Error: (%d) %s",
                      conf->dbi_driver, conf->dbi_user,
                      conf->dbi_host, conf->dbi_dbname, err_num, err_str);
        return APR_EGENERAL;
    }

    return APR_SUCCESS;
}
static apr_status_t
dbipool_destruct(void *sql, void *params, apr_pool_t * pool)
{
    dbi_conn_close((dbi_conn *) sql);
    return APR_SUCCESS;
}

static int
dbi_pool_post_config(apr_pool_t * p, apr_pool_t * plog,
                     apr_pool_t * ptemp, server_rec * s)
{
    int rval;
    apr_hash_index_t *idx;
    char *key;
    dbi_config *val;
    apr_ssize_t len;
    void *data;
    const char *userdata_key = "mod_dbi_pool_init";

    apr_pool_userdata_get(&data, userdata_key, s->process->pool);
    if (!data) {
        apr_pool_userdata_set((const void *) 1, userdata_key,
                              apr_pool_cleanup_null, s->process->pool);
        return OK;
    }

    ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p,
                  "[mod_dbi_pool.c] Running DBI Pool Init");
    if ((rval = dbi_initialize(NULL)) > 0) {
        ap_log_perror(APLOG_MARK, APLOG_DEBUG, 0, p,
                      "[mod_dbi_pool.c] Initialization of libdbi found %d drivers in default driver directory",
                      rval);
    }
    else {                      /* An error was returned or libdbi found 0 drivers */
        ap_log_perror(APLOG_MARK, APLOG_EMERG, 0, p,
                      "[mod_dbi_pool.c] - Initlialization of libdbi with default driver directory failed");
        return APR_EDSOOPEN;
    }

    /*
     * loop the hashed config stuff... 
     */
    for (idx = apr_hash_first(p, dbi_pool_config); idx;
         idx = apr_hash_next(idx)) {
        apr_hash_this(idx, (void *) &key, &len, (void *) &val);
#ifdef APR_HAS_THREADS
        if (apr_reslist_create(&val->pool, val->rec.conn_min,   /* hard minimum */
                               val->rec.conn_soft,      /* soft maximum */
                               val->rec.conn_max,       /* hard maximum */
                               val->rec.conn_ttl,       /* Time to live */
                               dbipool_construct,       /* Make a New Connection */
                               dbipool_destruct,        /* Kill Old Connection */
                               (void *) &val->rec, p) != APR_SUCCESS)
            ap_log_perror(APLOG_MARK, APLOG_EMERG, 0, p,
                          "[mod_dbi_pool.c] - Creation of dbi connection pool "
                          "failed for config set %s", key);
#else
        if (dbipool_construct(&val->conn, val->rec, p) != APR_SUCCESS) {
            ap_log_perror(APLOG_MARK, APLOG_EMERG, 0, p,
                          "[mod_dbi_pool.c] - Creation of DBI connection failed "
                          "for config set %s", key);
        }
#endif
        /*
         * XXX: If no new connection could be opened, the reslist will not
         * contain any resources and reslist_create will fail, giving 
         * val->pool the value NULL and making it unusable. The configuration 
         * set is not really recoverable in this case. For now we log this condition
         * and the authentication functions protect themselves against this 
         * condition. A much better way would be to free the configuration
         * set completely. But to do this other changes are necessary, so for
         * now this is basically a workaround. This has to be fixed before
         * 1.0 
         */
        apr_hash_set(dbi_pool_config, key, APR_HASH_KEY_STRING, val);
    }

    ap_add_version_component(p, "mod_dbi_pool/0.1.0");
    return OK;
}

static apr_status_t
dbi_pool_pre_config(apr_pool_t * pconf, apr_pool_t * plog, apr_pool_t * ptemp)
{
    apr_status_t rv = APR_SUCCESS;

    /*
     * create our globalish config var 
     */
    dbi_pool_config = apr_hash_make(pconf);
    return rv;
}

static void dbi_pool_hooks(apr_pool_t * p)
{
    ap_hook_pre_config(dbi_pool_pre_config, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_post_config(dbi_pool_post_config, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA dbi_pool_module = {
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    NULL,
    NULL,
    dbi_pool_cmds,
    dbi_pool_hooks
};

/* Functions we export for modules to use:
	- open acquires a connection from the pool (opens one if necessary)
	- close releases it back in to the pool
*/

dbi_conn *dbipool_open(conn_id id)
{
    dbi_conn *ret = NULL;
    dbi_config *conf = apr_hash_get(dbi_pool_config, id, APR_HASH_KEY_STRING);
#ifdef APR_HAS_THREADS
    if (apr_reslist_acquire(conf->pool, (void **) &ret) != APR_SUCCESS) {
        return NULL;
    }
    return ret;
#else
    return conf->conn;
#endif
}

void dbipool_close(conn_id id, dbi_conn * conn)
{
#ifdef APR_HAS_THREADS
    dbi_config *conf = apr_hash_get(dbi_pool_config, id, APR_HASH_KEY_STRING);
    apr_reslist_release(conf->pool, (void **) conn);
#endif
}
